class HandleSetListener:
    def __init__(self, app):
        self.app = app

    def render_message(self, message):
        # set handle on init
        if message.get("command") in ["console_response"]:
            if message.get("name") == "handle":
                self.app.handle = message.get("value")
            return


import bisect
from akris.log_config import LogConfig

logger = LogConfig.get_instance().get_logger("akris.timeline")

BEFORE = 0
AFTER = 1


class Annotation:
    def __init__(self, timestamp, content, position=AFTER):
        self.content = content
        self.timestamp = timestamp
        self.position = position


class TimestampGroup:
    def __init__(self):
        self.messages = []

    def add_message(self, new_message):
        # sort results with identical timestamps by message hash
        # this is kind of a heuristic because messages with identical timestamps are not
        # guaranteed to be descendents or antecedents of one another
        added = False
        for existing_message in self.messages:
            # if new_message is a descendent of message
            if existing_message.get("message_hash") in [
                new_message.get("net_chain"),
                new_message.get("self_chain"),
            ]:
                self.messages.append(new_message)
                added = True
                break

            # new_message is an antecedent of message
            if new_message["message_hash"] in [
                existing_message.get("net_chain"),
                existing_message.get("self_chain"),
            ]:
                self.messages.insert(self.messages.index(existing_message), new_message)
                added = True
                break
        if not added:
            self.messages.append(new_message)

    def count(self):
        message_count = len(self.messages)
        newline_count = 0
        multiline_header_count = 0
        for message in self.messages:
            newline_count += message.get("body").count("\n")

        # we must account for the multiline message headers as well
        for message in self.messages:
            if message.get("body").count("\n") > 0:
                multiline_header_count += 1

        return message_count + newline_count + multiline_header_count

    def newlines_before(self, message):
        message_index = self.messages.index(message)
        newline_count = 0
        for message in self.messages[:message_index]:
            newline_count += message.get("body").count("\n")
        return newline_count


class Timeline:
    def __init__(self):
        self.ordered_timestamps = []
        self.annotations = {}
        self.timestamp_groups = {}

    def get_line_count(self, message):
        if "\n" in message.get("body"):
            return 0

        # need to add one for the header
        return message.get("body").count("\n") + 1

    def calculate_message_index(self, message):
        ts = message.get("timestamp")
        tsg = self.get_timestamp_group(ts)

        index = 0
        index += self.antecedent_timestamp_group_count(ts)
        index += tsg.messages.index(message)
        index += tsg.newlines_before(message)
        index += self.count_annotations_before_ts_group(ts)

        # Text indexing starts at 1
        index += 1
        return index

    def calculate_annotation_index(self, annotation):
        ts = annotation.timestamp
        index = 0
        index += self.antecedent_timestamp_group_count(ts)
        index += self.count_earlier_annotations(ts)
        if annotation.position == AFTER:
            index += self.get_timestamp_group(ts).count()

        # Text indexing starts at 1
        index += 1
        return index

    def antecedent_timestamp_group_count(self, timestamp):
        count = 0
        for ts in self.timestamp_groups:
            if ts < timestamp:
                count += self.timestamp_groups[ts].count()
        return count

    def get_antecedent_timestamp(self, timestamp):
        previous_ts = None
        for ts in self.ordered_timestamps:
            if ts < timestamp:
                previous_ts = ts
        return previous_ts

    def get_decedent_timestamp(self, timestamp):
        decedent_ts = None
        for ts in reversed(self.ordered_timestamps):
            if ts > timestamp:
                decedent_ts = ts
        return decedent_ts

    def add_message(self, message):
        ts = message["timestamp"]
        if ts not in self.ordered_timestamps:
            bisect.insort(self.ordered_timestamps, ts)
        tsg = self.get_timestamp_group(ts)
        tsg.add_message(message)
        return self.calculate_message_index(message)

    def add_annotation(self, annotation):
        self.annotations[annotation.timestamp] = annotation
        return self.calculate_annotation_index(annotation)

    def remove_annotation(self, annotation):
        annotations_copy = self.annotations.copy()
        for ts in annotations_copy:
            if annotations_copy[ts].content == annotation.content:
                del self.annotations[ts]

    def annotated(self, timestamp):
        return self.annotations.get(timestamp)

    def get_annotation(self, content):
        for ts in self.annotations:
            if self.annotations[ts].content == content:
                return self.annotations[ts]

    def get_first_timestamp(self):
        return self.ordered_timestamps[0]

    def get_last_timestamp(self):
        return self.ordered_timestamps[-1]

    def get_timestamp_group(self, timestamp):
        if not self.timestamp_groups.get(timestamp):
            tsg = TimestampGroup()
            self.timestamp_groups[timestamp] = tsg
            return tsg
        else:
            return self.timestamp_groups[timestamp]

    def count_earlier_annotations(self, timestamp):
        count = 0
        for ts in self.annotations:
            if ts < timestamp:
                count += 1
        return count

    def count_annotations_before_ts_group(self, timestamp):
        count = 0
        for ts in self.annotations:
            if ts < timestamp:
                count += 1
        if self.annotations.get(timestamp):
            if self.annotations[timestamp].position == BEFORE:
                count += 1
        return count



import re
import tkinter as tk

from akris_desktop import emojis


def bytes_for_char_utf_8(c):
    return len(c.encode("utf-8"))


def count_bytes(s):
    b = 0
    total = 0
    for i, c in enumerate(s):
        cs = bytes_for_char_utf_8(c)
        total += cs
    return total


class MessageEntry(tk.Frame):
    def __init__(self, root, app):
        self.root = root
        self.app = app

        # Create a resizable pane with vertical orientation
        self.frame = tk.Frame(root)
        self.frame.pack(side=tk.BOTTOM, fill=tk.X)

        self.placehodler_text = "Write a message..."

        # Create a ScrolledText widget
        self.text = tk.Text(
            self.frame, wrap=tk.WORD, height=5, highlightthickness=0, padx=10, pady=10
        )
        self.text.configure(
            insertbackground="white", background="#333232", foreground="white"
        )
        self.text.pack(side=tk.LEFT, fill=tk.X, expand=True)

        # Add placeholder text
        self.text.insert(tk.END, self.placehodler_text)
        self.text.config(foreground="gray")

        self.context_menu = tk.Menu(root, tearoff=0)
        self.text.bind("<Button-3>", self.show_context_menu)
        self.configure_context_menu()

        # Bind events to the widget
        self.text.bind("<FocusIn>", self.on_entry_click)
        self.text.bind("<FocusOut>", self.on_focus_out)
        self.text.bind("<Return>", self.handle_enter)
        self.text.bind("<Tab>", self.autocomplete)
        self.text.bind("<Control-v>", self.handle_paste)

    def configure_context_menu(self):
        self.context_menu.add_command(
            label="Paste", command=lambda: self.handle_context_menu_paste()
        )
        self.context_menu.bind("<Leave>", lambda e: self.context_menu.unpost())

    def show_context_menu(self, event):
        self.context_menu.post(event.x_root, event.y_root)

    def handle_context_menu_paste(self):
        content = self.root.clipboard_get()
        clean_content = emojis.decode(content)
        self.text.insert(tk.INSERT, clean_content)

    def on_entry_click(self, event):
        if self.text.get("1.0", "end-1c") == self.placehodler_text:
            self.text.delete("1.0", tk.END)
            self.text.config(foreground="white")

    def on_focus_out(self, event):
        if self.text.get("1.0", "end-1c") == "":
            self.text.insert(tk.END, self.placehodler_text)
            self.text.config(foreground="#ECECEC")

    def ignore_enter(self, event):
        return "break"

    def handle_enter(self, event):
        if not (event.state & 1):  # Check if Shift is NOT pressed
            # Handle Enter key (send message)
            self.send_message()
        else:
            self.insert_new_line()
        return "break"

    def insert_new_line(self):
        # Insert a new line character at the current cursor position
        self.text.insert(tk.INSERT, "\n")

    def autocomplete(self, event) -> None:
        cursor_pos = self.text.index("insert")
        match = re.fullmatch(
            r"(.*\s)?([^\s:]+):? ?", self.text.get("0.0", self.text.index(tk.INSERT))
        )
        if match is None:
            return "break"
        preceding_text, last_word = match.groups()  # preceding_text can be None

        nicks = self.app.broadcast_tab.peers.get_peers()
        if last_word in nicks:
            completion = nicks[(nicks.index(last_word) + 1) % len(nicks)]
        else:
            try:
                completion = next(
                    username
                    for username in nicks
                    if username.lower().startswith(last_word.lower())
                )
            except StopIteration:
                return "break"

        if preceding_text:
            new_text = preceding_text + completion + " "
        else:
            new_text = completion + ": "
        self.text.delete("0.0", cursor_pos)
        self.text.insert("0.0", new_text)
        self.text.mark_set("insert", str(float(len(new_text))))
        return "break"

    def handle_paste(self, event):
        content = self.root.clipboard_get()
        clean_content = emojis.decode(content)
        self.text.insert(tk.INSERT, clean_content)
        return "break"

    def encode_action(self, message):
        pattern = r"\x01ACTION (.*?)\x01"
        if message.startswith("/me"):
            if len(message.encode("utf-8")) <= 325:
                pattern = r"^/me(.*)"
                match = re.search(pattern, message)
                if match:
                    predicate = match.group(1)
                    action_message = re.sub(pattern, f"\x01ACTION{predicate}\x01", message, count=1)
                    return action_message
        return message


import datetime
import re
import time
import tkinter as tk
from tkinter import ttk
import webbrowser
from functools import partial
from tkinter.font import Font
from akris.log_config import LogConfig

from .page_monitor import PageMonitor
from .text_utils import colorize
from .timeline import Timeline, Annotation, BEFORE

logger = LogConfig.get_instance().get_logger("akris_desktop.timeline_view")


# "unicode surrogates" seen in at least one message from phf - origin and purpose
# as yet unclear, but can't be rendered by the Text
def remove_unicode_surrogates(json_string):
    return re.sub(
        r"[\U0001F600-\U0001F64F\U0001F300-\U0001F5FF\U0001F680-\U0001F6FF\U0001F1E0-\U0001F1FF\u2600-\u26FF\u2700-\u27BF\U000024C2-\U0001F251]",
        "X",
        json_string,
    )


def indexize(index):
    return str(float(index))


def is_new_day(timestamp1, timestamp2):
    # Convert the timestamps to datetime objects
    dt1 = datetime.datetime.fromtimestamp(timestamp1)
    dt2 = datetime.datetime.fromtimestamp(timestamp2)

    # Return True if the dates are different, False otherwise
    return dt1.date() != dt2.date()


class DuplicateMessageException(Exception):
    pass


class CustomText(tk.Text):
    def __init__(self, *args, scroll_handler=None, **kwargs):
        self.scroll_handler = scroll_handler
        super().__init__(*args, **kwargs)

    def yview(self, *args):
        self.scroll_handler(*args)
        return super().yview(*args)


class MultipartMessage:
    def __init__(self, message):
        self.parts = [message]
        self.of = message.get("of")

    def add_part(self, message):
        for part in self.parts:
            if message.get("n") < part.get("n"):
                self.parts.insert(self.parts.index(part), message)
                return
        self.parts.append(message)

    def is_complete(self):
        return len(self.parts) == self.of

    def assembled_body(self):
        body = ""
        for part in self.parts:
            body += part.get("body")
        return body


class TimelineView(tk.Frame):
    def __init__(self, root, app):
        super().__init__(root)
        self.message_at_index = {}
        self.root = root
        self.app = app
        self.timeline = Timeline()
        self.frame = tk.Frame(root)
        self.frame.pack(fill=tk.BOTH, expand=True)
        self.text = CustomText(
            self.frame,
            wrap=tk.WORD,
            highlightthickness=0,
            padx=10,
            pady=10,
            scroll_handler=self.handle_scroll,
            borderwidth=0,
        )
        self.configure_scrollbar_style()
        self.scrollbar = ttk.Scrollbar(self.frame)
        self.scrollbar.pack(side=tk.RIGHT, fill=tk.Y)

        # Associate the Scrollbar with the Text widget
        self.text.config(yscrollcommand=self.scrollbar.set)
        self.scrollbar.config(command=self.text.yview)

        self.text.configure(state="disabled")
        self.text.pack(fill=tk.BOTH, expand=True)
        self.text.configure(
            background="black", foreground="white", font=("Courier New", 10)
        )
        self.context_menu = tk.Menu(root, tearoff=0)
        self.configure_context_menu()

        # tooltip display
        self.tooltip_label = None

        # configuration
        self.configure_tags()
        self.configure_bindings()

        self.character_insertion_index = 0

        # we need to keep track of this in order to determine when
        # to move to the end after receiving a new non page_response
        # message
        self.hit_bottom = True
        self.hit_top = False

        # paging
        self.filter = {}
        self.should_refresh_window = True
        self.page_monitor = PageMonitor()

        # for assembly of multipart messages
        self.multipart_staging = {}

    def show_tooltip(self, text):
        if self.tooltip_label:
            self.hide_tooltip()
        self.tooltip_label = tk.Label(self.root, fg="black", background="white", borderwidth=1, height=1, padx=0, pady=0)
        self.tooltip_label.place(x=0, y=0)
        self.tooltip_label.config(text=text, width=len(text))

    def hide_tooltip(self):
        if self.tooltip_label:
            self.tooltip_label.place_forget()
            self.tooltip_label.destroy()
            self.tooltip_label = None

    def await_message_stats(self):
        while not self.app.message_stats:
            self.root.after(100, self.await_message_stats)
            return
        self.page_monitor.days_before = self.minimum_days_before()
        self.tab.more_link.config(text=self.page_monitor.button_label())
        self.start_refresh_loop()

    def minimum_days_before(self):
        latest_ts = self.app.message_stats.get("latest_broadcast_message_timestamp")
        if not latest_ts:
            return 1
        latest_message_date = datetime.datetime.fromtimestamp(latest_ts)
        now = datetime.datetime.now()
        return (now - latest_message_date).days + 1

    def start_refresh_loop(self):
        if self.should_refresh_window:
            self.should_refresh_window = False
            self.load_page("page_up")

        self.root.after(500, self.start_refresh_loop)

    def render_message(self, message):
        if self.filter.get(message.get("message_hash")):
            raise DuplicateMessageException()

    def load_previous(self):
        self.hit_top = False
        self.page_monitor.days_before += 1
        self.tab.more_link.config(text=self.page_monitor.button_label())
        self.should_refresh_window = True

    def load_page(self, command):
        if command == "page_up":
            # Get the line number of the first visible line
            self.app.api_client.send_command(
                {
                    "command": command,
                    "args": [time.time(), self.page_monitor.days_before],
                }
            )

    def configure_scrollbar_style(self):
        style = ttk.Style()
        style.theme_use("default")
        style.configure(
            "TScrollbar",
            foreground="darkgray",
            troughcolor="black",
            background="darkgray",
            arrowcolor="white",
            bordercolor="black",
        )

    def configure_tags(self):
        lmargin = tk.font.Font(family="Courier New", size=11).measure(
            "[08:05:35]   asciilifeform | "
        )
        self.text.tag_configure(
            "speaker", font=Font(family="Courier New", weight="bold")
        )
        self.text.tag_configure(
            "text", font=Font(family="Courier New"), lmargin2=lmargin
        )
        self.text.tag_configure(
            "highlight_speaker", foreground="#E3ED1C", background="#D60D93"
        )
        self.text.tag_configure("ref_link", underline=True)
        self.text.tag_configure("url", underline=True)
        self.text.tag_configure("multiline_header", font=("", "11", "bold italic"))
        self.text.tag_configure("date_break", foreground="#1CEAED")
        self.text.tag_configure("timestamp", foreground="#B9BFC0")
        self.text.tag_configure("highlight", background="#D60D93")
        self.text.tag_configure("hearsay", foreground="#F6D12D")

    def configure_bindings(self):
        self.text.tag_bind(
            "ref_link", "<Button-1>", partial(self.on_ref_link_leftclick, "ref_link")
        )
        self.text.tag_bind(
            "ref_link", "<Enter>", (lambda e: self.text.config(cursor="hand2"))
        )
        self.text.tag_bind(
            "ref_link", "<Leave>", (lambda e: self.text.config(cursor="xterm"))
        )
        self.text.tag_bind("url", "<Button-1>", partial(self.on_link_leftclick, "url"))
        self.text.tag_bind(
            "url", "<Enter>", (lambda e: self.text.config(cursor="hand2"))
        )
        self.text.tag_bind(
            "url", "<Leave>", (lambda e: self.text.config(cursor="xterm"))
        )
        self.text.tag_bind(
            "timestamp", "<Enter>", (lambda e: self.text.config(cursor="hand2"))
        )
        self.text.tag_bind(
            "timestamp", "<Leave>", (lambda e: self.text.config(cursor="xterm"))
        )
        self.text.tag_bind(
            "multiline_header", "<Enter>", (lambda e: self.text.config(cursor="hand2"))
        )
        self.text.tag_bind(
            "multiline_header", "<Leave>", (lambda e: self.text.config(cursor="xterm"))
        )
        self.text.tag_bind("timestamp", "<Button-1>", (lambda e: self.copy_ref_link(e)))
        self.text.tag_bind("multiline_header", "<Button-1>", (lambda e: self.copy_ref_link(e)))
        self.text.bind("<Prior>", self.handle_page_up)

    def configure_context_menu(self):
        self.context_menu.add_command(
            label="Copy", command=lambda: self.copy_selected_text()
        )
        self.context_menu.bind("<Leave>", lambda e: self.context_menu.unpost())
        self.text.bind("<Button-3>", self.show_context_menu)

    def copy_ref_link(self, event):
        index = int(event.widget.index(f"@{event.x},{event.y}").split(".")[0])
        self.root.clipboard_clear()
        line = self.text.get(f"{index}.0", f"{index}.end")
        logger.info(line)
        message = self.message_at_index.get(index)
        self.root.clipboard_append(f"pest://{message.get('message_hash')}")

    def handle_scroll(*args):
        self = args[0]
        if len(args) == 3:
            if args[1] == "moveto":
                pos = self.text.yview()
                if pos[0] == 0.0:
                    if not self.hit_top:
                        self.hit_top = True
                if pos[1] == 1.0:
                    if not self.hit_bottom:
                        self.hit_bottom = True
                        self.hit_top = False
                else:
                    self.hit_bottom = False

    def get_insertion_index(self, line_index):
        return "{}.{}".format(line_index, self.character_insertion_index)

    def insert_text(self, text, line_index, tags=()):
        self.inserting_text = True
        self.text.configure(state="normal")
        self.text.insert(
            self.get_insertion_index(line_index),
            text,
            tags,
        )
        self.text.configure(state="disabled")
        self.character_insertion_index += len(text)

        def not_inserting():
            self.inserting_text = False

        self.root.after(500, not_inserting)

    def render_timestamp(self, timestamp, line_index, message):
        message_info_tag = f"message-info-{message.get('message_hash')}"
        tags = ["timestamp", message_info_tag]
        if message.get("hearsay"):
            tags.append("hearsay")
        msg_datetime = datetime.datetime.fromtimestamp(timestamp).strftime("%H:%M:%S")
        self.insert_text("[{}] ".format(msg_datetime), line_index, tags)
        self.bind_reporting_peers_tooltip(message_info_tag, message)

    def bind_reporting_peers_tooltip(self, message_info_tag, message):
        reporting_peers = message.get("reporting_peers")
        if reporting_peers:
            formatted_reporting_peers = ", ".join(reporting_peers)
            self.text.tag_bind(message_info_tag, "<Enter>", partial(self.on_timestamp_enter, formatted_reporting_peers))
            self.text.tag_bind(message_info_tag, "<Leave>", self.on_timestamp_leave)

    def render_speaker(self, message, is_action, line_index):
        if is_action:
            speaker = "*"
        else:
            speaker = message.get("speaker")
        body = message.get("body")
        tags = ["speaker"]
        justified_speaker = f"{speaker:>{15}}"
        if not is_action:
            self.text.tag_configure("colorize_" + speaker, foreground=colorize(speaker))
        if self.app.handle in body and not self.app.handle == message.get("speaker"):
            tags.append("highlight_speaker")
        else:
            tags.append("colorize_" + speaker)
        self.insert_text(
            justified_speaker,
            line_index,
            tags,
        )

    def render_divider(self, line_index):
        tags = []
        self.insert_text(
            " | ",
            line_index,
            tags,
        )

    def render_body(self, message, line_index):
        formatted_body = remove_unicode_surrogates(message.get("body"))
        self.insert_text(
            f"{formatted_body}\n",
            line_index,
            ("text",),
        )

    def extract_action_message(self, speaker, body):
        pattern = r"\x01ACTION (.*?)\x01"
        match = re.search(pattern, body)
        if match:
            return f"{speaker} {match.group(1)}"
        else:
            return body

    def render_annotation(self, annotation, line_index):
        self.character_insertion_index = 0
        self.insert_text(f"{annotation.content}\n", line_index, ("text", "date_break"))

    def render_multiline_header(self, message, line_index):
        self.character_insertion_index = 0
        ts = message.get("timestamp")
        msg_date = datetime.datetime.fromtimestamp(ts).strftime("%m/%d")
        msg_time = datetime.datetime.fromtimestamp(ts).strftime("%H:%M:%S")

        message_info_tag = f"message-info-{message.get('message_hash')}"
        tags = ["text", "multiline_header", message_info_tag]
        if message.get("hearsay"):
            tags.append("hearsay")
        self.insert_text(
            f"On {msg_date} at {msg_time} {message.get('speaker')} wrote:\n",
            line_index,
            tags,
        )
        reporting_peers = message.get("reporting_peers")
        if reporting_peers:
            formatted_reporting_peers = ", ".join(reporting_peers)
            self.text.tag_bind(message_info_tag, "<Enter>", partial(self.on_timestamp_enter, formatted_reporting_peers))
            self.text.tag_bind(message_info_tag, "<Leave>", self.on_timestamp_leave)

    def render_multiline(self, message, line_index):
        start = indexize(line_index + 1)
        body = message.get("body")
        lines = body.split("\n")
        for i, line in enumerate(lines):
            self.character_insertion_index = 0
            url_info = self.replace_bracket_pairs(line_index + i + 1, line)
            line_body = url_info.get("body")
            self.insert_text(f"{line_body}\n", line_index + i + 1, ("text",))
            self.tag_bracket_pairs(url_info)
        end = self.text.index(self.get_insertion_index(line_index + len(lines) + 1))
        self.find_and_tag_ref_links(self.text, start, end)
        self.find_and_tag_urls(self.text, start, end)

    def render_line(self, message, line_index):
        # initialize character insertion index
        self.character_insertion_index = 0
        is_action = False
        start = indexize(line_index)
        if "\x01" in message.get("body"):
            message["body"] = self.extract_action_message(
                message.get("speaker"), message.get("body")
            )
            is_action = True
        self.render_timestamp(message.get("timestamp"), line_index, message)
        self.render_speaker(message, is_action, line_index)
        self.render_divider(line_index)
        url_info = self.replace_bracket_pairs(line_index, message.get("body"))
        message["body"] = url_info.get("body")
        self.render_body(message, line_index)
        self.tag_bracket_pairs(url_info)

        end = self.text.index(self.get_insertion_index(line_index))

        self.find_and_tag_ref_links(self.text, start, end)
        self.find_and_tag_urls(self.text, start, end)
        self.reposition(message)

    def add_message(self, message):
        self.add_date_change(message.get("timestamp"))
        index = self.timeline.add_message(message)
        self.message_at_index[index] = message
        if "\n" in message.get("body"):
            self.render_multiline_header(message, index)
            self.render_multiline(message, index)
        else:
            self.render_line(message, index)

    def remove_message(self, message):
        logger.info("*********************** remove_message()")
        logger.info(message)
        index = self.timeline.calculate_message_index(message)
        line_count = self.timeline.get_line_count(message)
        self.text.configure(state="normal")
        self.text.delete(
            indexize(index),
            indexize(index + line_count),
        )
        self.text.configure(state="disabled")

    def add_date_change(self, timestamp):
        antecedent_timestamp = self.timeline.get_antecedent_timestamp(timestamp)
        decedent_timestamp = self.timeline.get_decedent_timestamp(timestamp)
        if antecedent_timestamp:
            if is_new_day(antecedent_timestamp, timestamp):
                self.insert_date_change(timestamp)
        if decedent_timestamp:
            if is_new_day(timestamp, decedent_timestamp):
                self.insert_date_change(decedent_timestamp)

    def insert_date_change(self, timestamp):
        if not self.timeline.annotated(timestamp):
            date = datetime.datetime.fromtimestamp(timestamp).strftime("%A, %B %d, %Y")
            annotation = Annotation(timestamp, f"--- {date} ---", BEFORE)
            self.remove_existing_annotation(annotation)
            annotation_index = self.timeline.add_annotation(annotation)
            self.render_annotation(annotation, annotation_index)

    def remove_existing_annotation(self, annotation):
        existing_annotation = self.timeline.get_annotation(annotation.content)
        if existing_annotation:
            existing_annotation_index = self.timeline.calculate_annotation_index(
                existing_annotation
            )
            self.delete_line(existing_annotation_index)
            self.timeline.remove_annotation(existing_annotation)

    def delete_line(self, line_index):
        self.text.configure(state="normal")
        start = f"{line_index}.0"
        end = f"{line_index + 1}.0"
        self.text.delete(start, end)
        self.text.configure(state="disabled")

    # handle positioning after receiving page responses
    def reposition(self, message):
        if not self.hit_bottom:
            return

        if self.hit_top:
            if not message.get("end"):
                return

        self.text.see("end")

    def find_and_tag_ref_links(self, textwidget: tk.Text, start: str, end: str) -> None:
        search_start = start
        while True:
            match_start = textwidget.search(
                r"\mpest://[a-zA-Z0-9=/+]", search_start, end, nocase=True, regexp=True
            )
            if not match_start:
                break

            url = textwidget.get(match_start, f"{match_start} lineend")

            url = url.split()[0]
            url = url.split("'")[0]
            url = url.split('"')[0]
            url = url.split("`")[0]

            # URL, and URL. URL? URL! (also URL). (also URL.)
            url = url.rstrip(".,?!")
            if "(" not in url:  # urls can contain spaces (e.g. wikipedia)
                url = url.rstrip(")")
            url = url.rstrip(".,?!")

            # [url][foobar]
            if "]" in url:
                pos = url.find("]")
                if pos < url.find("["):
                    url = url[:pos]

            match_end = f"{match_start} + {len(url)} chars"
            textwidget.tag_add("ref_link", match_start, match_end)
            search_start = f"{match_end} + 1 char"

    def replace_bracket_pairs(self, line_index, body):
        positions = []

        def url_replacer(m):
            url = m.group(1)
            text = m.group(2)

            positions.append(
                {
                    "url": url,
                    "match_start": f"{line_index}.{self.character_insertion_index + match.start()}",
                    "match_end": f"{line_index}.{self.character_insertion_index + match.start() + len(text)}",
                    "tag_name": f"url-{url}",
                }
            )
            return text

        pattern = r'\[([^\[\]]+?)\]\[([^\[\]]+?)\]'
        match = re.search(pattern, body)
        while (match):
            body = re.sub(pattern, url_replacer, body, count=1)
            match = re.search(pattern, body)
        return {
            "body": body,
            "positions": positions
        }

    def tag_bracket_pairs(self, url_info):
        positions = url_info.get("positions")
        for position in positions:
            self.text.tag_add("url", position.get("match_start"), position.get("match_end"))
            self.text.tag_add(position.get("tag_name"), position.get("match_start"), position.get("match_end"))
            if position.get("url").startswith("pest://"):
                self.text.tag_bind(position.get("tag_name"), "<Button-1>",
                                   partial(self.on_named_ref_link_click, position.get("url")))
            else:
                self.text.tag_bind(position.get("tag_name"), "<Button-1>",
                                   partial(self.on_named_link_click, position.get("url")))
            self.text.tag_bind(position.get("tag_name"), "<Enter>", partial(self.on_link_enter, position.get("url")))
            self.text.tag_bind(position.get("tag_name"), "<Leave>", self.on_link_leave)

    def find_and_tag_urls(self, textwidget: tk.Text, start: str, end: str) -> None:
        search_start = start
        while True:
            match_start = textwidget.search(
                r"\mhttps?://[a-z0-9:]", search_start, end, nocase=True, regexp=True
            )
            if not match_start:  # empty string means not found
                break

            url = textwidget.get(match_start, f"{match_start} lineend")

            url = url.split()[0]
            url = url.split("'")[0]
            url = url.split('"')[0]
            url = url.split("`")[0]

            # URL, and URL. URL? URL! (also URL). (also URL.)
            url = url.rstrip(".,?!")
            if "(" not in url:  # urls can contain spaces (e.g. wikipedia)
                url = url.rstrip(")")
            url = url.rstrip(".,?!")

            # [url][foobar]
            if "]" in url:
                pos = url.find("]")
                if pos < url.find("["):
                    url = url[:pos]

            match_end = f"{match_start} + {len(url)} chars"
            textwidget.tag_add("url", match_start, match_end)
            search_start = f"{match_end} + 1 char"

    def on_ref_link_leftclick(self, tag, event):
        tag_range = event.widget.tag_prevrange(tag, "current + 1 char")
        assert tag_range
        start, end = tag_range
        text = event.widget.get(start, end)
        message_hash = text.split("pest://")[1]
        self.app.api_client.send_command(
            {"command": "page_around", "args": [message_hash]}
        )

    def on_named_ref_link_click(self, url, event):
        message_hash = url.split("pest://")[1]
        self.app.api_client.send_command(
            {"command": "page_around", "args": [message_hash]}
        )

    def on_named_link_click(self, url, event):
        webbrowser.open(url)

    def on_link_leftclick(self, tag, event):
        # To test this, set up 3 URLs, and try clicking first and last char of middle URL.
        # That finds bugs where it finds the wrong URL, or only works in the middle of URL, etc.
        tag_range = event.widget.tag_prevrange(tag, "current + 1 char")
        assert tag_range
        start, end = tag_range
        text = event.widget.get(start, end)
        webbrowser.open(text)

    def copy_selected_text(self):
        # Get the selected text
        selected_text = self.text.get("sel.first", "sel.last")

        # Copy the selected text to the clipboard
        self.text.clipboard_clear()
        self.text.clipboard_append(selected_text)

    def show_context_menu(self, event):
        self.context_menu.post(event.x_root, event.y_root)

    def handle_page_up(self, event):
        self.text.yview_scroll(-1, "pages")
        if self.text.yview()[0] == 0.0:
            pass

    def highlight_line(self, line_index):
        self.text.tag_add("highlight", f"{line_index}.0", f"{line_index + 1}.0")

    def on_link_enter(self, url, event):
        self.show_tooltip(url)

    def on_link_leave(self, event):
        self.hide_tooltip()

    def on_timestamp_enter(self, reporting_peers, event):
        self.show_tooltip(reporting_peers)

    def on_timestamp_leave(self, event):
        self.hide_tooltip()

    def is_hearsay(self, message):
        if message.get("immediate"):
            return False

        if message.get("speaker") == self.app.handle:
            return False

        return True



import tkinter as tk


class Notifications(tk.Frame):
    def __init__(self, root, app, timeline_view):
        super().__init__(root, bg="black", borderwidth=0, height=30, pady=10, padx=0)
        self.app = app
        self.root = root
        self.timeline_view = timeline_view
        self.load_more_button = None

    def display_load_previous(self):
        if not self.load_more_button:
            self.load_more_button = tk.Button(
                self,
                text="more...",
                command=self.timeline_view.load_previous,
                height=1,
                pady=0,
                padx=0,
            )
            self.load_more_button.pack(side=tk.RIGHT)

    def clear(self):
        if self.load_more_button:
            self.load_more_button.destroy()
            self.load_more_button = None



import hashlib
import random
import sqlite3
import json
import string
import time

import pytest
import pytest_mock
from akris_desktop.app import App
from akris.client import Client
import akris.lib.simple_graph_sqlite.database as db
from unittest.mock import MagicMock
from queue import Queue
import tkinter as tk

def generate_random_string(length):
    """Helper function to generate random strings of given length."""
    return ''.join(random.choices(string.ascii_letters + string.digits, k=length))


def generate_chat_messages(num_messages):
    """Generate a list of num_messages random chat messages."""
    messages = []
    prev_hash = ''
    for i in range(num_messages):
        timestamp = int(time.time())
        body = generate_random_string(20)
        hash = hashlib.sha256(body.encode()).hexdigest()
        handle = generate_random_string(10)
        netchain = prev_hash if prev_hash else ''
        message = {
            'command': 'broadcast_text',
            'timestamp': timestamp,
            'body': body,
            'message_hash': hash,
            'speaker': handle,
            'net_chain': netchain
        }
        messages.append(message)
        prev_hash = hash
    return messages

@pytest.fixture
def client():
    akris_path = '/home/awt/PycharmProjects/pine-beetle/tests/fixtures/akris.db'
    db.initialize(akris_path)
    connection = sqlite3.connect(akris_path)
    cursor = connection.cursor()
    latest_message = json.loads(cursor.execute("SELECT * FROM nodes ORDER BY json_extract(body, '$.timestamp') DESC LIMIT 1;").fetchone()[0])
    messages = db.traverse(akris_path, latest_message['message_hash'],
                                    neighbors_fn=db.find_inbound_neighbors, with_bodies=True)
    parsed_messages = []
    for message_row in messages:
        message = json.loads(message_row[2])
        if not message:
            pass
        else:
            message["command"] = "broadcast_text"
            parsed_messages.append(message)

    # parsed_messages = generate_chat_messages(1000)
    client = MagicMock(Client)
    client.message_queue = Queue()
    for message in parsed_messages:
        client.message_queue.put(message)

    yield client

    cursor.close()
    connection.close()

@pytest.fixture
def app(client):
    root = tk.Tk()
    app = App(root, client)
    yield app
    # root.destroy()


def test_display_broadcast_text(app):
    app.wait_visibility()
    app.check_message_queue()
    app.update()
    # assert(len(app.message_table.table.get_children()) == 40)
    app.mainloop()

def test_console_command(app):
    app.wait_visibility()
    app.check_message_queue()
    app.notebook.select(1)
    app.console.entry.focus_force()
    app.console.entry.insert(0, 'at')
    app.update()
    app.console.entry.event_generate('<Return>')
    app.update()
    assert(app.api_client.send_command.called_with({
        'command': 'at',
        'args': []
    }))
    app.mainloop()



import re
import tkinter as tk
from functools import partial
from ..timeline_view import remove_unicode_surrogates

class SearchResult:
    def __init__(self, record):
        (
            self.speaker,
            self.recipient_handle,
            self.body,
            self.base64_message_hash
         ) = record

class Search:
    def __init__(self, output, api_client):
        self.output = output
        self.search_results = {}
        self.api_client = api_client
        self.output.tag_bind("message_hash", "<Button-1>", partial(self.on_message_hash_click, "message_hash"))
        self.output.tag_bind(
            "message_hash", "<Enter>", (lambda e: self.output.config(cursor="hand2"))
        )
        self.output.tag_bind(
            "message_hash", "<Leave>", (lambda e: self.output.config(cursor="xterm"))
        )

    def render(self, search_results):
        if len(search_results.get("body")) == 0:
            self.insert_text("No results\n")

        for search_result in search_results.get("body"):
            self.render_search_result(SearchResult(search_result))

    def on_message_hash_click(self, tag, event):
        tag_range = event.widget.tag_prevrange(tag, "current + 1 char")
        assert tag_range
        start, end = tag_range
        message_hash = event.widget.get(start, end).rstrip()
        search_result = self.search_results[message_hash]
        if search_result.recipient_handle:
            args = [message_hash, search_result.recipient_handle]
        else:
            args = [message_hash]
        self.api_client.send_command(
            {"command": "page_around", "args": args}
        )

    def render_search_result(self, search_result):
        self.search_results[search_result.base64_message_hash] = search_result
        self.output.character_insertion_index = 0
        self.render_message_hash(search_result.base64_message_hash)
        self.render_speaker(search_result.speaker)
        self.render_body(search_result.body)
        # line_index = int(self.output.index(tk.END).split(".")[0])
        # self.output.tag_add("error", f"{int(line_index)-2}.0", self.output.index(tk.END))

    def render_message_hash(self, message_hash):
        self.insert_text(
            f"{message_hash}",
            ("message_hash",),
        )

    def render_speaker(self, speaker):
        justified_speaker = f"{speaker:>{20}}"
        self.insert_text(
            f"{justified_speaker} | ",
        )

    def render_body(self, body):
        body = remove_unicode_surrogates(body)
        positions = self.replace_search_terms(body)
        self.insert_text(
            f"{positions.get('body')}\n",
            ("body",),
        )
        for position in positions["positions"]:
            self.tag_search_terms(position)

    def tag_search_terms(self, position):
        self.output.tag_add(
            "search_term",
            position["match_start"],
            position["match_end"],
        )

    def replace_search_terms(self, body):
        line_index = int(self.output.index(tk.END).split(".")[0]) - 1
        positions = []
        pattern = r'<<([^<>]+)>>'
        match = re.search(pattern, body)
        while match:
            body = re.sub(pattern, r'\1', body, count=1)
            positions.append({
                "match_start": f"{line_index}.{self.output.character_insertion_index + match.start()}",
                "match_end": f"{line_index}.{self.output.character_insertion_index + match.start() + len(match.group(1))}",
            })
            match = re.search(pattern, body)
        return {
            "body": body,
            "positions": positions,
        }

    def insert_text(self, text, tags=()):
        self.output.configure(state="normal")
        line_index = self.output.index(tk.END).split(".")[0]
        self.output.insert(
            self.get_insertion_index(line_index),
            text,
            tags,
        )
        self.output.configure(state="disabled")
        self.output.character_insertion_index += len(text)

    def get_insertion_index(self, line_index):
        return "{}.{}".format(line_index, self.output.character_insertion_index)



from akris.log_config import LogConfig

logger = LogConfig.get_instance().get_logger("akris_desktop.broadcast_view")

from .timeline_view import TimelineView, MultipartMessage


class BroadcastView(TimelineView):
    def __init__(self, root, app, tab):
        super().__init__(root, app)
        self.tab = tab
        self.root = root

    def render_message(self, message):
        super().render_message(message)

        if message.get("ref_link_page_response"):
            return

        if message.get("command") == "message_update":
            # if we already have this message we need to remove it and update it when the page refresh arrives
            if self.page_monitor.in_window(message):
                if message.get("message_hash") in self.filter:
                    self.remove_message(message)
                    del self.filter[message.get("message_hash")]
                    self.should_refresh_window = True

        if message["command"] not in ["broadcast_text", "broadcast_text_m"]:
            return

        # if this is a realtime message or get_data response, check if we should refresh the window
        if not message.get("page_response") and not message.get(
            "ref_link_page_response"
        ):
            if self.page_monitor.in_window(message):
                self.should_refresh_window = True
            return

        if message.get("command") == "broadcast_text_m":
            self.filter[message.get("message_hash")] = True
            if not message.get("text_hash") in self.multipart_staging:
                self.multipart_staging[message.get("text_hash")] = MultipartMessage(
                    message
                )
            else:
                self.multipart_staging[message.get("text_hash")].add_part(message)

            multipart_message = self.multipart_staging[message.get("text_hash")]
            if multipart_message.is_complete():
                message["body"] = multipart_message.assembled_body()
                del self.multipart_staging[message.get("text_hash")]
            else:
                return

        if self.is_hearsay(message):
            message["hearsay"] = True

        # track only messages that are page responses
        self.filter[message.get("message_hash")] = True
        self.add_message(message)



import argparse
import os

from datetime import datetime

import requests

GEMOJI_RELEASE_URL = "https://api.github.com/repos/github/gemoji/releases"
GEMOJI_JSON_DB_URL = (
    "https://raw.githubusercontent.com/github/gemoji/{tag}/db/emoji.json"
)


def get_lastest_release():
    req = requests.get(GEMOJI_RELEASE_URL)
    req.raise_for_status()

    data = req.json()

    latest = data[0]

    return latest["tag_name"], latest["name"]


def generate(path, dbname):
    tag, name = get_lastest_release()

    req = requests.get(GEMOJI_JSON_DB_URL.format(tag=tag))
    req.raise_for_status()

    data = req.json()

    path = os.path.join(path, dbname)

    with open(path, "w", encoding="utf-8") as file:
        file.write("### This is a generated file.\n")
        file.write("### Do not edit this file.\n")
        file.write("### Date: {0}\n".format(datetime.now().isoformat()[:-7]))
        file.write("### This file is based on {0}.\n".format(name))
        file.write("\n")
        file.write("from collections import namedtuple\n")
        file.write("\n")
        file.write(
            'Emoji = namedtuple("Emoji", ["aliases", "emoji", "tags", "category", "unicode_version"])\n'
        )
        file.write("\n")
        file.write("EMOJI_DB = [\n")

        for emoji in data:
            if "emoji" in emoji:
                file.write(
                    '    Emoji({aliases}, "{emoji}", {tags}, "{category}", "{unicode_version}"),\n'.format(
                        **{
                            "aliases": emoji["aliases"],
                            "emoji": emoji["emoji"],
                            "tags": emoji["tags"],
                            "category": emoji["category"],
                            "unicode_version": emoji["unicode_version"],
                        }
                    )
                )

        file.write("]\n")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Generates the Emoji database.")
    parser.add_argument("--dir", default=".", help="Database location")
    parser.add_argument("--dbname", default="db.py", help="Database location")
    args = parser.parse_args()

    generate(args.dir, args.dbname)



import tkinter as tk
from .message_entry import MessageEntry, count_bytes


class BroadcastMessageEntry(MessageEntry):
    def __init__(self, root, app):
        super().__init__(root, app)

    def handle_slash_command(self, message):
        pass

    def send_message(self):
        # Get the text from the text entry widget
        message = self.text.get("1.0", tk.END).rstrip("\n")
        if message[0] == "/":
            message = self.encode_action(message)

        # Clear the text entry widget
        self.text.delete("1.0", tk.END)

        if not self.app.disable_multipart:
            # count the number of bytes in the message body
            # if necessary send a multipart message
            if count_bytes(message) > 324:
                self.app.api_client.send_command(
                    {"command": "broadcast_text_m", "args": [message]}
                )
                return

        self.app.api_client.send_command(
            {"command": "broadcast_text", "args": [message]}
        )



import tkinter as tk


class At:
    def __init__(self, output):
        self.output = output

    def render(self, response):
        peers = response.get("body")
        if len(peers) == 0:
            self.render_no_peers()
        else:
            for peer in response.get("body"):
                self.render_peer(peer)

    def render_peer(self, peer):
        self.output.configure(state="normal")
        self.output.insert(
            tk.END,
            "{}\t{}\t{}\n".format(
                peer.get("handle"), peer.get("address"), peer.get("active_at")
            ),
        )
        self.output.configure(state="disabled")

    def render_no_peers(self):
        self.output.configure(state="normal")
        self.output.insert(tk.END, "No peers found.\n")
        self.output.configure(state="disabled")



import tkinter as tk


class VersionInfo:
    def __init__(self, output):
        self.output = output

    def render(self, response):
        akris_version = response.get("version")
        pest_version = response.get("pest_version")
        earliest_supported_pest_version = response.get(
            "earliest_supported_pest_version"
        )
        body = (
            "Welcome to the Akris Desktop Pest console!\n\n"
            f"Akris station version: {akris_version}\n"
            f"Pest version: {hex(pest_version)}\n"
            f"Minimum pest version: {hex(earliest_supported_pest_version)}\n\n"
            f"Enter a pest command below or type help for a list of supported commands.\n\n"
        )

        self.output.configure(state="normal")
        self.output.insert(tk.END, body)
        self.output.configure(state="disabled")



import importlib
import logging
import os
import tkinter as tk
from tkinter import ttk
from PIL import Image, ImageTk

from akris_desktop.broadcast_tab import BroadcastTab
from akris_desktop.direct_tab import DirectTab
from akris_desktop.ref_link_tab import RefLinkTab
from akris_desktop.console import Console
from akris_desktop.message_stats_listener import MessageStatsListener
from akris_desktop.new_handle_listener import NewHandleListener
from akris_desktop.new_ref_link_listener import NewRefLinkListener
from akris_desktop.handle_set_listener import HandleSetListener
from akris_desktop.timeline_view import DuplicateMessageException

logger = logging.getLogger("akris_desktop")


class App(tk.Frame):
    def __init__(self, root=None, api_client=None, options=None):
        super().__init__(root)
        try:
            with importlib.resources.path("akris_desktop", "images") as resource_path:
                ico = Image.open(os.path.join(resource_path, "locust_icon.png"))
                photo = ImageTk.PhotoImage(ico)
                root.wm_iconphoto(False, photo)
        except FileNotFoundError:
            logger.exception("icon file missing.")
        self.disable_multipart = options.disable_multipart
        self.remote_station = options.remote_station
        self.handle = None
        self.root = root
        self.root.title("Akris")
        self.root.configure(bg="black")
        self.configure(bg="black")
        self.api_client = api_client
        self.message_detail = None
        self.new_handle_listener = NewHandleListener(self)
        self.new_ref_link_listener = NewRefLinkListener(self)
        self.handle_set_listener = HandleSetListener(self)
        self.message_listeners = []
        self.direct_tabs = []
        self.ref_link_tabs = []
        self.message_stats = None
        self.pack(fill=tk.BOTH, expand=True)
        root.protocol("WM_DELETE_WINDOW", self.on_close)

        # create notebook
        style = ttk.Style()
        style.configure("TNotebook", background="#333232", bordercolor="black")
        style.configure("TNotebook.Tab", background="#2C2A2A", foreground="darkgray")
        style.map(
            "TNotebook.Tab",
            background=[("selected", "#333232")],
            foreground=[("selected", "white")],
        )
        self.notebook = ttk.Notebook(self, style="TNotebook")
        self.notebook.pack(fill=tk.BOTH, expand=True)

        # create main tab
        self.broadcast_tab = BroadcastTab(self.root, self)
        self.broadcast_tab.configure(bg="black")
        self.notebook.add(self.broadcast_tab, text="broadcast messages")
        self.console_tab = tk.Frame(self.root)
        self.console_tab.configure(bg="black")
        self.notebook.add(self.console_tab, text="console")

        # setup console tab
        self.console_frame = tk.Frame(self.console_tab)
        self.console_frame.pack(fill=tk.BOTH, expand=True)
        self.console = Console(self.console_frame, self.api_client)

        # register message listeners
        self.register_message_listener(self.console)
        self.register_message_listener(self.new_handle_listener)
        self.register_message_listener(self.new_ref_link_listener)
        self.register_message_listener(self.handle_set_listener)
        self.register_message_listener(MessageStatsListener(self))
        self.api_client.send_command({"command": "message_stats", "args": []})
        self.api_client.send_command({"command": "version_info", "args": []})
        self.api_client.send_command({"command": "knob", "args": ["handle"]})
        self.api_client.send_command({"command": "report_presence", "args": []})

    def add_direct_message_tab(self, handle):
        # check if tab already exists
        for tab in self.direct_tabs:
            if tab.handle == handle:
                self.notebook.select(tab)
                return
        dm_tab = DirectTab(self.root, handle, self)
        self.direct_tabs.append(dm_tab)
        self.notebook.insert(1, dm_tab, text=handle)
        return dm_tab

    def add_ref_link_tab(self, message_hash):
        # check if tab already exists
        for tab in self.ref_link_tabs:
            if tab.message_hash == message_hash:
                self.notebook.select(tab)
                return
        ref_link_tab = RefLinkTab(self.root, message_hash, self)
        self.ref_link_tabs.append(ref_link_tab)
        self.notebook.insert(1, ref_link_tab, text=ref_link_tab.title())
        self.notebook.select(ref_link_tab)
        return ref_link_tab

    def open_direct_message_tab(self, handle):
        tab = self.add_direct_message_tab(handle)
        self.notebook.select(tab)

    def close_ref_link_tab(self, tab):
        self.ref_link_tabs.remove(tab)
        self.notebook.forget(tab)
        self.unregister_message_listener(tab.message_table)
        self.notebook.select(self.broadcast_tab)

    def close_tab(self, tab):
        self.direct_tabs.remove(tab)
        self.notebook.forget(tab)
        self.unregister_message_listener(tab.message_table)
        self.notebook.select(self.broadcast_tab)

    def register_message_listener(self, listener):
        self.message_listeners.append(listener)

    def unregister_message_listener(self, listener):
        self.message_listeners.remove(listener)

    def check_message_queue(self):
        # check the queue and perform actions if needed
        if not self.api_client.message_queue.empty():
            while not self.api_client.message_queue.empty():
                message = self.api_client.message_queue.get()
                for listener in self.message_listeners:
                    try:
                        listener.render_message(message)
                    except DuplicateMessageException:
                        pass

    def run(self):
        self.check_message_queue()
        self.root.after(
            100, self.run
        )  # schedule the function to be called again in 1 second

    def on_close(self):
        logger.info("closing")
        self.root.destroy()
        self.root.quit()
        if not self.remote_station:
            self.api_client.shutdown_station()
            return

        self.api_client.disconnect()



import datetime
import time

from .timeline_view import TimelineView


class DirectView(TimelineView):
    def __init__(self, root, handle, app, tab):
        self.handle = handle
        self.tab = tab
        super().__init__(root, app)

    def minimum_days_before(self):
        latest_ts = self.app.message_stats.get("direct_message_timestamps", {}).get(
            self.handle
        )
        if not latest_ts:
            return 1
        latest_message_date = datetime.datetime.fromtimestamp(latest_ts)
        now = datetime.datetime.now()
        return (now - latest_message_date).days + 1

    def render_message(self, message):
        # check for duplicates
        super().render_message(message)

        if message.get("ref_link_page_response"):
            return

        # is this a direct_text?
        if message.get("command") not in ["direct_text", "direct_text_m"]:
            return

        # is this the right dm tab?
        if not (
            message.get("handle") == self.handle
            or message.get("speaker") == self.handle
        ):
            return

        # if this is a realtime message or get_data response, update the window if it's in range
        self.update_window(message)

        # is this a realtime message or get_data response?
        if not message.get("page_response"):
            return

        # add message to filter and render
        self.filter[message.get("message_hash")] = True
        self.add_message(message)

    def load_page(self, command):
        if command == "page_up":
            self.app.api_client.send_command(
                {
                    "command": command,
                    "args": [time.time(), self.page_monitor.days_before, self.handle],
                }
            )

    def update_window(self, message):
        if not message.get("page_response"):
            if self.page_monitor.in_window(message):
                self.should_refresh_window = True



import time

import pytest
from akris_desktop.app import App
from akris.client import Client
from unittest.mock import MagicMock
from queue import Queue
import tkinter as tk


@pytest.fixture
def client():
    client = MagicMock(Client)
    client.message_queue = Queue()
    timestamp = int(time.time())
    messages = [
        {
            'command': 'console_response',
            'type': 'at',
            'body': [
                {
                    'handle': 'atahualpa',
                    'address': '122.122.0.1:8081',
                    'active_at': '2023-05-21 07:13:26.541438',
                    'active_at_unixtime': 1684678406
                }
            ]
        }
    ]
    for message in messages:
        client.message_queue.put(message)

    yield client

@pytest.fixture
def app(client):
    root = tk.Tk()
    app = App(root, client)
    app.wait_visibility()
    app.check_message_queue()
    app.notebook.select(1)
    yield app
    # root.destroy()


def test_at_command(app):
    app.console.entry.focus_force()
    app.console.entry.insert(0, 'at')
    app.update()
    app.console.entry.event_generate('<Return>')
    app.update()
    assert(app.api_client.send_command.called_with({
        'command': 'at',
        'args': []
    }))
    app.mainloop()

def test_at_response_display(app):
    app.update()
    app.mainloop()



import tkinter as tk


class Knob:
    def __init__(self, output):
        self.output = output

    def render(self, response):
        knobs = response.get("body")
        for key, value in knobs.items():
            self.render_knob("{}:\t{}".format(key, value))

    def render_knob(self, knob):
        self.output.configure(state="normal")
        self.output.insert(tk.END, "{}\n".format(knob))
        self.output.configure(state="disabled")



import tkinter as tk


class Default:
    def __init__(self, output):
        self.output = output
        pass

    def render(self, response):
        if response.get("body"):
            self.output.configure(state="normal")
            self.output.insert(tk.END, "{}\n".format(response.get("body")))
            self.output.configure(state="disabled")



class MessageStatsListener:
    def __init__(self, app):
        self.app = app

    def render_message(self, message):
        if message.get("command") in ["console_response"]:
            if message.get("type") == "message_stats":
                self.app.message_stats = {
                    "latest_broadcast_message_timestamp": message.get(
                        "latest_broadcast_message_timestamp"
                    ),
                    "direct_message_timestamps": message.get(
                        "direct_message_timestamps"
                    ),
                }



#!/usr/bin/env python3

import time
import os
import logging
import tkinter as tk
import argparse

PEST_HOST = "0.0.0.0"
PEST_PORT = 8080
API_HOST = "127.0.0.1"
API_PORT = 8081
AKRIS_DATA_PATH = os.path.join(os.path.expanduser("~"), ".akris")
AKRIS_DATABASE_NAME = "akris.db"


def get_options():
    parser = argparse.ArgumentParser()
    parser.add_argument("--pest-host", default=PEST_HOST, help="pest station host")
    parser.add_argument("--api-host", default=API_HOST, help="api service host")
    parser.add_argument("--pest-port", default=PEST_PORT, type=int, help="pest station port")
    parser.add_argument("--api-port", default=API_PORT, type=int, help="api service port")
    parser.add_argument("--akris-data-path", default=AKRIS_DATA_PATH, help="path at which to locate persistent Akris data")
    parser.add_argument("--akris-database-name", default=AKRIS_DATABASE_NAME, help="name to use for the Akris database file")
    parser.add_argument("--disable-multipart", action='store_true', default=False, help="disable sending multipart messages")
    parser.add_argument("--remote-station", action='store_true', default=False, help="connect to a remote station rather than spinning up a local one")
    return parser.parse_args()


def main():
    """Main entry point for akris-desktop command."""
    options = get_options()

    from akris.log_config import LogConfig
    LogConfig(data_path=options.akris_data_path).get_logger("akris_desktop.main")
    from akris.station import Station
    from akris import client
    from akris_desktop.app import App

    logging.basicConfig(level=logging.INFO)

    if not options.remote_station:
        station = Station(
            tcp_host=options.api_host,
            tcp_port=options.api_port,
            udp_port=options.pest_port,
            host=options.pest_host,
            data_path=options.akris_data_path,
            database_name=options.akris_database_name,
        )
        station.start()
        while not station.ready():
            time.sleep(0.1)

    c = client.Client(host=options.api_host, port=int(options.api_port))
    root = tk.Tk()
    root.configure(bg="black")
    app = App(root, c, options=options)
    app.run()
    root.mainloop()
    if not options.remote_station:
        station.stop()


if __name__ == '__main__':
    main()


import tkinter as tk

from .direct_message_entry import DirectMessageEntry
from .direct_view import DirectView


class DirectTab(tk.Frame):
    def __init__(self, root, handle, app):
        super().__init__(root)
        self.handle = handle
        self.app = app
        self.toolbar_frame = tk.Frame(self, background="#333232")
        self.toolbar_frame.pack(side=tk.TOP, fill=tk.X)
        self.close_link = tk.Label(
            self.toolbar_frame,
            text="🗙",
            fg="white",
            cursor="hand2",
            background="red",
        )
        self.close_link.pack(side=tk.RIGHT)
        self.close_link.bind("<Button-1>", lambda e: self.app.close_tab(self))
        self.messages_and_peers_frame = tk.Frame(self)
        self.messages_and_peers_frame.pack(fill=tk.BOTH, expand=True)
        self.message_table = DirectView(
            self.messages_and_peers_frame, handle, app, self
        )
        self.more_link = tk.Label(
            self.toolbar_frame,
            text=self.message_table.page_monitor.button_label(),
            fg="#00a1ff",
            cursor="hand2",
            background="#333232",
        )
        self.message_table.await_message_stats()
        self.more_link.pack(side=tk.LEFT)
        self.more_link.bind("<Button-1>", lambda e: self.message_table.load_previous())
        self.message_entry = DirectMessageEntry(self, app, handle)
        self.app.register_message_listener(self.message_table)



import tkinter as tk


class Error:
    def __init__(self, output):
        self.output = output

    def render(self, response):
        error = response.get("body")
        self.render_error(error)

    def render_error(self, error):
        self.output.configure(state="normal")
        self.output.tag_remove("sel", "1.0", tk.END)
        self.output.mark_set(tk.INSERT, tk.END)
        start = self.output.index(tk.INSERT)
        self.output.insert(tk.END, "{}\n".format(error))
        end = "{}.end".format(int(float(start)))
        self.output.tag_add("error", start, end)
        self.output.configure(state="disabled")



import datetime
import tkinter as tk
import random
import webbrowser
from functools import partial
from tkinter.font import Font


class Master(tk.Frame):
    def __init__(self, root, app):
        self.app = app
        self.chain = app.timeline
        self.frame = tk.Frame(root)
        self.frame.pack(fill=tk.BOTH, expand=True)

        self.table = tk.Text(
            self.frame, wrap=tk.WORD, highlightthickness=0, padx=10, pady=10
        )
        self.table.configure(state="disabled")
        self.table.tag_configure("speaker", font=Font(family="Courier New"))
        self.table.tag_config("text", font=Font(family="Courier New"), lmargin2=260)
        self.table.pack(fill=tk.BOTH, expand=True)
        self.table.configure(background="black", foreground="white")
        self.table.tag_configure("url", underline=True)
        self.table.tag_bind(
            "url", "<Button-1>", partial(self._on_link_leftclick, "url")
        )
        self.table.tag_bind(
            "url", "<Enter>", (lambda e: self.table.config(cursor="hand2"))
        )
        self.table.tag_bind(
            "url", "<Leave>", (lambda e: self.table.config(cursor="xterm"))
        )
        self.context_menu = tk.Menu(root, tearoff=0)
        self.context_menu.add_command(
            label="Copy", command=lambda: self.copy_selected_text()
        )
        self.context_menu.bind("<Leave>", lambda e: self.context_menu.unpost())
        self.table.bind("<Button-3>", self.show_context_menu)

    def insert(self, message, index):
        msg_datetime = datetime.datetime.fromtimestamp(message["timestamp"]).strftime(
            "%H:%M:%S"
        )
        self.table.configure(state="normal")
        start = index
        message["index"] = str(index)
        datetime_str = "[{}] ".format(msg_datetime)
        self.table.insert(index, datetime_str)
        speaker = f"{message['speaker']:>{15}}"
        self.table.tag_configure(
            "colorize_" + speaker, foreground=self.colorize(speaker)
        )
        self.table.insert(
            "{}.{}".format(int(float(index)), len(datetime_str)),
            speaker,
            ("speaker", "colorize_" + speaker),
        )
        body = "{}".format(
            ": " + message["body"],
        )
        self.table.insert(
            "{}.{}".format(int(float(index)), len(datetime_str) + len(speaker)),
            "{}{}".format(body, "\n"),
            ("text",),
        )
        end = self.table.index(tk.INSERT)
        self.table.configure(state="disabled")
        self.find_and_tag_urls(self.table, start, end)
        self.table.see(tk.END)

    def render_message(self, message):
        if message["command"] == "broadcast_text":
            # self.chain.add_message(message)
            # index = self.chain.index(message["message_hash"])
            # self.insert(message, index)
            pass

    def colorize(self, speaker):
        # Set the seed based on the peer name
        random.seed(speaker)

        # Generate random RGB values for lighter colors
        r = random.randint(0, 255)
        g = random.randint(0, 255)
        b = random.randint(0, 255)

        # Generate the color code
        color_code = "#{:02X}{:02X}{:02X}".format(r, g, b)

        return color_code

    def find_and_tag_urls(self, textwidget: tk.Text, start: str, end: str) -> None:
        search_start = start
        while True:
            match_start = textwidget.search(
                r"\mhttps?://[a-z0-9:]", search_start, end, nocase=True, regexp=True
            )
            if not match_start:  # empty string means not found
                break

            url = textwidget.get(match_start, f"{match_start} lineend")

            url = url.split()[0]
            url = url.split("'")[0]
            url = url.split('"')[0]
            url = url.split("`")[0]

            # URL, and URL. URL? URL! (also URL). (also URL.)
            url = url.rstrip(".,?!")
            if "(" not in url:  # urls can contain spaces (e.g. wikipedia)
                url = url.rstrip(")")
            url = url.rstrip(".,?!")

            # [url][foobar]
            if "]" in url:
                pos = url.find("]")
                if pos < url.find("["):
                    url = url[:pos]

            match_end = f"{match_start} + {len(url)} chars"
            textwidget.tag_add("url", match_start, match_end)
            search_start = f"{match_end} + 1 char"

    def _on_link_leftclick(self, tag, event):
        # To test this, set up 3 URLs, and try clicking first and last char of middle URL.
        # That finds bugs where it finds the wrong URL, or only works in the middle of URL, etc.
        tag_range = event.widget.tag_prevrange(tag, "current + 1 char")
        assert tag_range
        start, end = tag_range
        text = event.widget.get(start, end)
        webbrowser.open(text)

    def copy_selected_text(self):
        # Get the selected text
        selected_text = self.table.get("sel.first", "sel.last")

        # Copy the selected text to the clipboard
        self.table.clipboard_clear()
        self.table.clipboard_append(selected_text)

    def show_context_menu(self, event):
        self.context_menu.post(event.x_root, event.y_root)



import tkinter as tk


class Wot:
    def __init__(self, output):
        self.output = output

    def render(self, response):
        peers = response.get("body")
        if len(peers) == 0:
            self.render_no_peers()
        else:
            for peer in response.get("body"):
                self.render_peer(peer)

    def render_peer(self, peer):
        if peer.get("address") and peer.get("port"):
            address = "{}:{}".format(peer.get("address"), peer.get("port"))
        else:
            address = "<address not configured>"
        self.output.configure(state="normal")
        self.output.insert(
            tk.END, "{}\t{}\n".format(" ".join(peer.get("handles")), address)
        )
        if len(peer.get("keys")) > 0:
            for key in peer.get("keys"):
                self.output.insert(tk.END, "\t{}\n".format(key))
        self.output.configure(state="disabled")

    def render_no_peers(self):
        self.output.configure(state="normal")
        self.output.insert(tk.END, "No peers found.\n")
        self.output.configure(state="disabled")



import importlib
import os
import tkinter as tk
from tkinter import ttk

from PIL import Image, ImageTk
from akris_desktop.console_renderers.at import At
from akris_desktop.console_renderers.default import Default
from akris_desktop.console_renderers.error import Error
from akris_desktop.console_renderers.help import Help
from akris_desktop.console_renderers.knob import Knob
from akris_desktop.console_renderers.search import Search
from akris_desktop.console_renderers.version_info import VersionInfo
from akris_desktop.console_renderers.wot import Wot


class Console(tk.Frame):
    def __init__(self, root, api_client):
        super().__init__(root)
        self.root = root
        self.history = []
        self.history_index = 0
        self.frame = tk.Frame(root)
        self.frame.pack(fill=tk.BOTH, expand=True)
        self.api_client = api_client
        self.output = tk.Text(
            self.frame, wrap=tk.WORD, highlightthickness=0, padx=10, pady=10
        )
        self.configure_scrollbar_style()
        self.scrollbar = ttk.Scrollbar(self.frame)
        self.scrollbar.pack(side=tk.RIGHT, fill=tk.Y)
        self.output.configure(yscrollcommand=self.scrollbar.set)
        self.scrollbar.configure(command=self.output.yview)
        self.output.character_insertion_index = 0
        self.output.configure(state="disabled")
        self.output.configure(background="black", foreground="green")
        self.output.pack(fill=tk.BOTH, expand=True)

        # Setup the context menus
        self.output_context_menu = tk.Menu(self.output, tearoff=0)
        self.configure_output_context_menu()

        self.entry = tk.Entry(self.frame, highlightthickness=0)
        self.entry.configure(
            insertbackground="green", background="black", foreground="green"
        )
        self.entry.pack(fill=tk.X)

        # Setup the entry context menus
        self.entry_context_menu = tk.Menu(self.output, tearoff=0)
        self.configure_entry_context_menu()


        # Bind events to the widget
        self.entry.bind("<Return>", self.handle_enter)
        self.entry.bind("<Up>", self.handle_up_arrow)
        self.entry.bind("<Down>", self.handle_down_arrow)

        # Setup tags:
        self.output.tag_configure("error", foreground="red")
        self.output.tag_configure("search_term", font=tk.font.Font(size=11, weight="bold"))
        lmargin2 = tk.font.Font(size=11).measure("nGsDUOHFRCmP0mG1h/KCDOSLM3hY4KBxthI2XF6B7D4=              awt | ")
        self.output.tag_configure("body", lmargin2=lmargin2)

        try:
            with importlib.resources.path("akris_desktop", "images") as images_path:
                self.insert_image(os.path.join(images_path, "pixelated_green_flag.png"))
            self.insert_message({"body": "\n\n"})
        except FileNotFoundError:
            pass

        # Setup renderers:
        self.renderers = {
            "at": At(self.output),
            "error": Error(self.output),
            "help": Help(self.output),
            "knob": Knob(self.output),
            "search": Search(self.output, self.api_client),
            "version_info": VersionInfo(self.output),
            "wot": Wot(self.output),
        }

    def render_message(self, message):
        if message["command"] == "console_response":
            self.renderers.get(message.get("type"), Default(self.output)).render(
                message
            )
            self.output.see(tk.END)

    def configure_entry_context_menu(self):
        self.entry_context_menu.add_command(
            label="Paste", command=lambda: self.handle_entry_context_menu_paste()
        )
        self.entry_context_menu.bind("<Leave>", lambda e: self.entry_context_menu.unpost())
        self.entry.bind("<Button-3>", self.show_entry_context_menu)

    def show_entry_context_menu(self, event):
        self.entry_context_menu.post(event.x_root, event.y_root)

    def handle_entry_context_menu_paste(self):
        content = self.root.clipboard_get()
        self.entry.insert(tk.INSERT, content)

    def configure_output_context_menu(self):
        self.output_context_menu.add_command(
            label="Copy", command=lambda: self.copy_output_selected_text()
        )
        self.output_context_menu.bind("<Leave>", lambda e: self.output_context_menu.unpost())
        self.output.bind("<Button-3>", self.show_output_context_menu)

    def copy_output_selected_text(self):
        # Get the selected text
        selected_text = self.output.get("sel.first", "sel.last")

        # Copy the selected text to the clipboard
        self.output.clipboard_clear()
        self.output.clipboard_append(selected_text)

    def show_output_context_menu(self, event):
        self.output_context_menu.post(event.x_root, event.y_root)

    def configure_scrollbar_style(self):
        style = ttk.Style()
        style.theme_use("default")
        style.configure(
            "TScrollbar",
            foreground="darkgray",
            troughcolor="black",
            background="darkgray",
            arrowcolor="white",
            bordercolor="black",
        )
    def insert_message(self, message):
        self.output.configure(state="normal")
        self.output.insert(tk.END, message["body"])
        self.output.configure(state="disabled")

    def insert_image(self, image_path):
        image = Image.open(image_path)
        photo = ImageTk.PhotoImage(image)
        self.output.image_create("insert", image=photo)
        self.output.image = photo


    def handle_enter(self, event):
        if event.state == 0:
            # Handle Enter key (send message)
            self.execute_command()
        return "break"

    def handle_up_arrow(self, event):
        # if the index is 0, do nothing
        # Get the clipboard content and insert it into the Text widget
        if len(self.history) > 0:
            self.entry.delete("0", tk.END)
            self.entry.insert("0", self.history[self.history_index])
            if self.history_index > 0:
                self.history_index -= 1

    def handle_down_arrow(self, event):
        # Get the clipboard content and insert it into the Text widget
        self.entry.delete("0", tk.END)
        if self.history_index < len(self.history) - 1:
            self.history_index += 1
            self.entry.insert("0", self.history[self.history_index])

    def execute_command(self):
        # Get the text from the text entry widget
        command_string = self.entry.get()
        args = command_string.split(" ")[1:]
        command = command_string.split(" ")[0]
        if command == "search":
            args = [" ".join(args)]
        if command == "searchs":
            args = command_string.split(" ")[2:]
            speaker = command_string.split(" ")[1]
            args = [speaker, " ".join(args)]

        # Log the command to the command history
        self.history.append(command_string)
        self.history_index = len(self.history) - 1

        # Clear the text entry widget
        self.entry.delete("0", tk.END)

        self.insert_message({"body": "> " + command_string + "\n"})
        self.api_client.send_command({"command": command, "args": args})

    def configure_scrollbar_style(self):
        pass



import tkinter as tk

from .ref_link_view import RefLinkView


class RefLinkTab(tk.Frame):
    def __init__(self, root, message_hash, app):
        super().__init__(root)
        self.toolbar_frame = tk.Frame(self, background="#333232")
        self.close_link = tk.Label(
            self.toolbar_frame,
            text="🗙",
            fg="white",
            cursor="hand2",
            background="red",
        )
        self.close_link.pack(side=tk.RIGHT)
        self.close_link.bind("<Button-1>", lambda e: self.app.close_ref_link_tab(self))
        # self.close_button = ttk.Button(self.toolbar_frame, text="close", command=lambda: self.app.close_ref_link_tab(self))
        # self.close_button.pack(side=tk.RIGHT)
        self.toolbar_frame.pack(side=tk.TOP, fill=tk.X)
        self.message_hash = message_hash
        self.app = app
        self.messages_frame = tk.Frame(self)
        self.messages_frame.pack(fill=tk.BOTH, expand=True)
        self.message_table = RefLinkView(self.messages_frame, app, message_hash, self)
        self.app.register_message_listener(self.message_table)

    def title(self):
        title = self.message_table.title
        if not title:
            return self.message_hash[0:15]
        return title



class NewRefLinkListener:
    def __init__(self, app):
        self.app = app

    def render_message(self, message):
        if not message.get("ref_link_page_response"):
            return

        if message.get("ref_link_hash") not in [
            self.app.notebook.tab(t, "text") for t in self.app.ref_link_tabs
        ]:
            self.app.add_ref_link_tab(message.get("ref_link_hash"))



from .timeline_view import TimelineView, MultipartMessage


class RefLinkView(TimelineView):
    def __init__(self, root, app, message_hash, tab):
        super().__init__(root, app)
        self.tab = tab
        self.title = None
        self.message_hash = message_hash
        self.text.tag_configure("reflink_highlight", foreground="black", background="#FAF182")

    def render_message(self, message):
        super().render_message(message)

        if not message.get("ref_link_page_response"):
            return

        if self.message_hash != message.get("ref_link_hash"):
            return

        if message.get("command") == "broadcast_text_m":
            self.filter[message.get("message_hash")] = True
            if not message.get("text_hash") in self.multipart_staging:
                self.multipart_staging[message.get("text_hash")] = MultipartMessage(
                    message
                )
            else:
                self.multipart_staging[message.get("text_hash")].add_part(message)

            multipart_message = self.multipart_staging[message.get("text_hash")]
            if multipart_message.is_complete():
                message["body"] = multipart_message.assembled_body()
                del self.multipart_staging[message.get("text_hash")]
            else:
                return

        # set the title from the linked message
        if message.get("message_hash") == self.message_hash:
            self.title = message.get("body")[0:15]
            self.app.notebook.tab(self.tab, text=self.title)

        if self.is_hearsay(message):
            message["hearsay"] = True

        self.filter[message.get("message_hash")] = True
        self.add_message(message)

        if message.get("end"):
            self.text.see(f"{self.highlighed_line_index}.0")

    def highlight_line(self, line_index):
        self.text.tag_add("reflink_highlight", f"{line_index}.0", f"{line_index + 1}.0")
        self.text.see(f"{line_index}.0")

    def add_message(self, message):
        super().add_message(message)
        if message.get("message_hash") == self.message_hash:
            line_index = self.timeline.calculate_message_index(message)
            self.highlighed_line_index = line_index
            self.highlight_line(line_index)
